import os
import struct
import sys

import numpy as np
from PyQt5.QtCore import Qt, QSettings, QDir, QFileInfo
from PyQt5.QtWidgets import QMainWindow, QShortcut, QApplication, QFileDialog, QTableWidgetItem

import pySPM
from pySPM.tools.spectraviewer import Ui_SpectraViewer

DPI = 100.0


class SpectraViewer(QMainWindow):
    def __init__(self, filename=None, parent=None):
        super(SpectraViewer, self).__init__(parent)
        self.ui = Ui_SpectraViewer()
        self.ui.setupUi(self)
        self.sf = 7200
        self.k0 = 0
        self.dsf = 0
        self.dk0 = 0
        self.ita = None
        self.fig = self.ui.mpl.canvas.fig
        self.canvas = self.ui.mpl.canvas
        self.ax = self.fig.add_subplot(111)
        self.nextMass = QShortcut(Qt.Key_Plus, self)
        self.prevMass = QShortcut(Qt.Key_Minus, self)
        self.nextMass.activated.connect(self.next_mass)
        self.prevMass.activated.connect(self.prev_mass)
        self.ui.pushButton_2.clicked.connect(self.toggleMassCal)
        self.ui.pushButton.clicked.connect(self.removeMassCalItem)
        self.ui.show_mass.clicked.connect(self.yAxisScaleChanged)
        self.canvas.mpl_connect('motion_notify_event', self.on_motion)
        self.canvas.mpl_connect('button_press_event', self.onMousePress)
        self.canvas.mpl_connect('button_release_event', self.onMouseRelease)
        self.canvas.mpl_connect("scroll_event", self.scrolling)
        self.labels = []
        self.action = None
        self.lab_lines = []
        self.MassCal = []
        self.open(filename)

    def resizeEvent(self, event):
        return
        sx, sy = self.ui.fig.width() / DPI, self.ui.fig.height() / DPI
        self.fig.set_size_inches(sx, sy)
        self.ui.fig.updateGeometry()
        self.canvas.updateGeometry()
        self.canvas.draw()
        self.canvas.flush_events()

    def __del__(self):
        if self.ita is not None:
            del self.ita
        self.ita = None

    def removeMassCalItem(self):
        row = self.ui.tableMassCal.selectedItems()[0].row()
        del self.MassCal[row]
        self.DoMassCal()

    def refresh(self):
        r = self.ax.get_xlim()
        self.yAxisScaleChanged()
        self.canvas.draw()
        self.canvas.flush_events()

    def next_mass(self):
        r = self.ax.get_xlim()
        self.ax.set_xlim(r[0] + 1, r[1] + 1)
        self.refresh()

    def prev_mass(self):
        r = self.ax.get_xlim()
        self.ax.set_xlim(r[0] - 1, r[1] - 1)
        self.refresh()

    def clear_labels(self):
        for x in self.labels:
            x.remove()
        self.labels[:] = []
        for x in self.lab_lines:
            self.ax.lines.remove(x)
        self.lab_lines[:] = []

    def plot_labels(self, colors=['r', 'g', 'b']):
        r = self.ax.get_xlim()
        E = []
        for nm in range(int(np.round(r[0], 0)), int(np.round(r[1], 0)) + 1):
            E += pySPM.utils.get_peaklist(nm, self.ita.polarity == 'Negative')
        m0s = [pySPM.utils.get_mass(x) for x in E]
        P = list(zip(m0s, E))
        P.sort(key=lambda x: x[0])
        y = self.ax.get_ylim()[1]
        for i, (mi, Ei) in enumerate(P):
            dmi = 2 * np.sqrt(mi) * np.sqrt((self.dk0 ** 2 / (self.sf ** 2)) + mi * (self.dsf ** 2 / (self.sf ** 2)))
            col = colors[i % len(colors)]
            self.lab_lines.append(self.ax.axvline(mi, color=col, alpha=.5))
            self.labels.append(
                self.ax.annotate(Ei, (mi, y), (5, 0), rotation=90, va='top', ha='left', textcoords='offset pixels'))
            if dmi > 0:
                self.lab_lines.append(self.ax.axvline(mi + dmi, color=col, alpha=.5, ls=':'))
                self.lab_lines.append(self.ax.axvline(mi - dmi, color=col, alpha=.5, ls=':'))

    def yAxisScaleChanged(self):
        r = self.ax.get_xlim()
        delta = r[1] - r[0]
        self.clear_labels()

        if self.ita is not None:
            SatLevel = self.ita.size['pixels']['x'] * self.ita.size['pixels']['y'] * self.ita.Nscan
            self.sat_level.set_ydata(SatLevel)

        max = 0
        left = int(pySPM.utils.mass2time(r[0], sf=self.sf, k0=self.k0) / 2)
        right = int(pySPM.utils.mass2time(r[1], sf=self.sf, k0=self.k0) / 2) + 1
        if left < 0:
            left = 0
        if right >= len(self.S):
            right = len(self.S) - 1
        if left < self.t[-1] and right > 0:
            max = np.max(self.S[left:right + 1])
            if max > 0:
                self.ax.set_ylim(0, 1.2 * max)
        if delta < 10:
            self.ui.show_mass.setEnabled(True)
            if self.ui.show_mass.isChecked():
                self.plot_labels()
        else:
            self.ui.show_mass.setEnabled(False)
        m0 = pySPM.utils.time2mass(left + right, self.sf, self.k0)
        dm = 2 * np.sqrt(m0) * np.sqrt((self.dk0 ** 2 / (self.sf ** 2)) + m0 * (self.dsf ** 2 / (self.sf ** 2)))
        self.ui.lab_m0.setText("m0 = {:.5f} ± {:.5f}".format(m0, dm))

    def scrolling(self, event):
        r = self.ax.get_xlim()
        delta = (r[1] - r[0])
        m0 = event.xdata
        zfact = 2
        if event.button == "down":
            zfact = 1 / zfact
        low = m0 - (m0 - r[0]) * zfact
        high = m0 + (r[1] - m0) * zfact
        self.ax.set_xlim((low, high))
        self.refresh()

    def open(self, t_filename=None):
        settings = QSettings(QSettings.IniFormat, QSettings.UserScope, "pySPM", "pySPM")
        if t_filename is None:
            home = QDir.cleanPath(os.getenv("HOMEPATH"))
            path = settings.value("lastPath", home)
            self.filename = QFileDialog.getOpenFileName(None, "Choose measurement file", path, "*.ita")[0]
        else:
            self.filename = t_filename
        check_file = QFileInfo(self.filename)
        self.setWindowTitle(check_file.fileName())
        if not check_file.exists() or not check_file.isFile():
            return

        settings.setValue("lastPath", check_file.path())
        self.ita = pySPM.ITA(self.filename, readonly=False)
        self.t, self.S = self.ita.getSpectrum(time=True)
        self.sf, self.k0 = self.ita.get_mass_cal()
        self.mass = pySPM.utils.time2mass(self.t, self.sf, self.k0)
        self.spec = self.ax.plot(self.mass, self.S)[0]
        SatLevel = self.ita.size['pixels']['x'] * self.ita.size['pixels']['y'] * self.ita.Nscan
        self.sat_level = self.ax.axhline(SatLevel, color='r')
        if 'pySPM' in self.ita.root.goto("MassScale"):
            self.MassCal = []
            N = self.ita.root.goto("MassScale/pySPM/N").get_ulong()
            for i in range(N):
                elt = self.ita.root.goto("MassScale/pySPM/" + str(i) + "/elt").value.decode('utf8')
                mass = self.ita.root.goto("MassScale/pySPM/" + str(i) + "/mass").get_double()
                time = self.ita.root.goto("MassScale/pySPM/" + str(i) + "/time").get_double()
                self.MassCal.append(dict(elt=elt, mass=mass, time=time))
        else:
            self.MassCal = []
            for x in self.ita.root.goto("MassScale/calib"):
                if x.name == 'assign':
                    self.MassCal.append({'elt': x.get_string()})
                if x.name == 'mcp':
                    mcp = struct.unpack("<10d", x.value)
                    self.MassCal[-1]['time'] = mcp[0]
                    self.MassCal[-1]['mass'] = mcp[1]
        self.DoMassCal()

    def get_mass(self, formula):
        if self.ita is not None and not (formula.endsWith('+') or formula.endsWith('-')):
            if self.ita.polarity == 'Negative':
                pol = "-"
            else:
                pol = "+"
            formula = formula + pol
        return pySPM.utils.get_mass(formula)

    def plotSpectra(self):
        self.mass = pySPM.utils.time2mass(self.t, self.sf, self.k0)
        self.spec.set_xdata(self.mass)
        self.refresh()

    def onMousePress(self, event):
        if event.button == 3:
            x = event.xdata

            i = pySPM.utils.closest_arg(self.mass, x)
            last = i - 1;
            while i != last:
                last = i
                i = i - 10 + np.argmax(self.S[i - 10:i + 10])
            I = self.S[i]
            self.MassCal.append(dict(time=self.t[i]))
            print("clicked @{}u (t={})".format(x, self.t[i]))
        elif event.button == Qt.LeftButton:
            self.action = ('move', event.xdata)
        else:
            print(event)

    def on_motion(self, event):
        r = self.ax.get_xlim()
        if self.action is not None:
            if self.action[0] == 'move':
                if event.xdata is not None:
                    delta = r[1] - r[0]
                    dx = event.xdata - self.action[1]
                    self.ax.set_xlim((r[0] - dx, r[1] - dx))
                    self.refresh()

    def onMouseRelease(self, event):
        if event.button == 3:
            x = event.xdata
            r = self.ax.get_xlim()
            frags = []
            for nm in range(int(np.round(r[0], 0)), int(np.round(r[1], 0)) + 1):
                frags += pySPM.utils.get_peaklist(nm, self.ita.polarity == 'Negative')
            masses = np.array([pySPM.utils.get_mass(x) for x in frags])
            dm = masses - x
            i = np.argmin(np.abs(dm))
            self.MassCal[-1]['mass'] = masses[i]
            self.MassCal[-1]['elt'] = frags[i]
            print("assigned to {}".format(frags[i]))
            self.DoMassCal()
        self.action = None
        self.refresh()

    def save_mass_cal(self):
        root = self.ita.root.goto("MassScale")
        for i, mcp in enumerate(self.MassCal):
            for key in mcp:
                if key in ['elt']:
                    data = mcp[key].encode('utf8') + b'\x00' * (256 - len(mcp[key]))
                    root.edit_block("pySPM/" + str(i), key, data)
                else:
                    root.edit_block("pySPM/" + str(i), key, struct.pack("<d", mcp[key]))
        root.edit_block("pySPM", "N", struct.pack("<I", len(self.MassCal)))

    def DoMassCal(self):
        ts = [x['time'] for x in self.MassCal]
        ms = [x['mass'] for x in self.MassCal]
        if len(ts) > 1:
            self.sf, self.k0, self.dsf, self.dk0 = pySPM.utils.fitSpectrum(ts, ms, error=True)
        else:
            self.k0 = ts[0] - self.sf * np.sqrt(ms[0])
            dsf = 0
            dk0 = 0
        self.ui.lab_k0.setText("k0 = {} ± {}".format(self.k0, self.dk0))
        self.ui.lab_sf.setText("sf = {} ± {}".format(self.sf, self.dsf))

        self.ita.setK0(self.k0)
        self.ita.setSF(self.sf)

        self.ui.tableMassCal.clearContents()
        self.ui.tableMassCal.setRowCount(len(self.MassCal))
        for i in range(len(self.MassCal)):
            self.ui.tableMassCal.setItem(i, 0, QTableWidgetItem(self.MassCal[i]['elt']))
            m = pySPM.utils.time2mass(self.MassCal[i]['time'], self.sf, self.k0)
            self.ui.tableMassCal.setItem(i, 1, QTableWidgetItem("{:.3f}".format(self.MassCal[i]['mass'])))
            self.ui.tableMassCal.setItem(i, 2, QTableWidgetItem("{:.0f}".format(self.MassCal[i]['time'])))
            delta = "{:.6f}".format(m - self.MassCal[i]['mass'])
            self.ui.tableMassCal.setItem(i, 3, QTableWidgetItem(delta))
        self.mass = pySPM.utils.time2mass(self.t, self.sf, self.k0)
        self.save_mass_cal()
        self.plotSpectra()

    def toggleMassCal(self):
        vis = not self.ui.tableMassCal.isVisible()
        self.ui.tableMassCal.setVisible(vis)
        self.ui.pushButton.setVisible(vis)
        if vis:
            self.ui.pushButton_2.setText("«")
        else:
            self.ui.pushButton_2.setText("»")

    def closeEvent(self, event):
        print("Closing spectra tool...")
        del self.ita


def main(filename=None):
    if filename is None and len(sys.argv) > 1:
        filename = sys.argv[1]
    print("Loading file \"{}\"".format(filename))
    app = QApplication(sys.argv)
    window = SpectraViewer(filename)
    window.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



import sys

cmd = None

if len(sys.argv) > 1:
    cmd = sys.argv[1]

commands = {}


def run_script(func):
    import importlib
    mod = importlib.import_module("pySPM.tools." + func)
    sys.argv = [sys.argv[0]] + [x for x in sys.argv[2:]]
    mod.main()


cmds = {
    'stability': dict(doc="Run a stability plotter tool in a GUI", script='stability'),
    'plotter': dict(doc="Display a live graphical display of all the logged parameters (usually the emission current)",
                    alias=['emission_current_plotter', 'parameters_plotter'], script='emission_current_plotter'),
    'timer': dict(
        doc="Display informations about the current measurement time. In particular this function estimate the remaining time from the measurement proportion and the elapsed time.",
        alias=['tof_timer', 'measurement_timer'], script='tof_timer'),
    'spectra': dict(doc="Spectra plotter and mass calibration GUI", script="spectra")
}

for c in cmds:
    doc = cmds[c].get('doc', '')
    script = cmds[c].get('script', None)
    commands[c] = (script, doc)
    if 'alias' in cmds[c]:
        for a in cmds[c]['alias']:
            commands[a] = (script, doc)

if cmd is None or cmd in ['help', '--help', '-h', '?', '-?']:
    print("Give as argument the command that pySPM should run. Valid commands are:")
    for c in cmds:
        print("{}: {} (aliases: {})".format(c, cmds[c].get('doc', ''), ",".join(cmds[c].get('alias', []))))
elif cmd in commands:
    print("Running command \"{}\"".format(cmd))
    run_script(commands[cmd][0])
else:
    print("Command \"{}\" not understood".format(cmd))
    print("Valid commands are:")
    for c in commands:
        print("{}: {}".format(c, commands[c][1]))



# -- coding: utf-8 --

# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

"""
Helper functions to handle spectras.
"""

from .misc import dec_debug, do_debug, alias


def get_substance_peaks(substance, negative=True):
    import os
    import sqlite3
    DB_PATH = os.path.join(os.path.abspath(os.path.join(__file__, "../..")), "data", "elements.db")
    conn = sqlite3.connect(DB_PATH)
    c = conn.cursor()
    c.execute(
        "SELECT Peaks.Fragment from Peaks where Peaks.Substance==(SELECT ID from substance where Name LIKE '%{name}%') and Polarity{pol}=0".format(
            name=substance, pol='><'[negative]))
    return [x[0] for x in c.fetchall()]


def get_dm(m, sf, k0, dsf, dk0):
    import numpy as np
    return 2 * np.sqrt(m) * np.sqrt((dk0 ** 2 / (sf ** 2)) + m * (dsf ** 2 / (sf ** 2)))


@alias("showPeak")
def show_peak(m, D, m0, delta=None, errors=False, dm0=0, dofit=False, show_elts=True,
              debug=False, Aredux=1, label=None, include=None, include_only=None, exclude=[],
              polarity="+", colors='rgb', pretty=True, formula=True, auto_scale=True, fakefit=False, zero_axis=True,
              **kargs):
    """
    Plot a spectrum (given by its mass vector m and intensity vector D) around a mass m0 (Δm=delta).
    
    Parameters
    ----------
    m : numpy.ndarray
        mass vector
    D : numpy.ndarray
        Intensity vector
    m0 : float
        Central mass of the plot's x-axis
    delta: None or float
        the half-span of the plot's x-axis.
        If None will try to determine automatically the peaks region
    errors : boolean
        If true will include the fitting errors (if dofit is True)
    dofit : boolean
        If True will fit the peaks for the elements found
    show_elts : boolean
        If True will display a list of all elements for the plotted region. The basic elements comes from a database, but they can be adjusted with the include, include_only and exclude parameters.
    include : list
        List of elements to be included to the plotting elements
    include_only : list
        Only those elements will be displayed and none from the database
    exclude : list
        Exclude some elements from display
    fakefit : boolean
        If the fit fails, fakefit can be used in order to see and adjust the first fitting guess
    dm0 : float
        The mass shift used for the fitting first guess.
    Aredux : float
        The amplitude reduction used to the first fitting guess.
    label : string
        Plot label. Will be displayed if you use the legend() function from matplotlib.
    polarity : string ("+" or "-")
        will switch the elements chosen from the database
    colors : string
        A series of letters defining matplotlib color used to display the elements.
    pretty : boolean
        If True will display the elements in a pretty way such that the elements text are not overlapping (or at least will avoid it as much as possible)
    formula : boolean
        Will convert the formula in a pretty manner. Exponents and indices will be written correctly
    auto_scale : boolean
        If True will try to rescale the plot in order to have all labels inside.
    zero_axis : boolean or dict
        Will draw an horizontal line for the zero intensities.
        If a dict, the parameters of the lines can be tuned. See the parameters passed to axhline from matplotlib
    """
    from . import LG, get_mass, get_peaklist
    from .elts import formulafy
    from scipy.optimize import curve_fit
    import numpy as np
    import copy
    import matplotlib.pyplot as plt

    if "showElts" in kargs:
        from warnings import warn
        warn("Parameter showElts is deprecated. Please use show_elts")
        show_elts = kargs.pop("showElts")

    if include is None:
        include = []

    if type(include) is str:
        include = include.split(',')

    assert type(include) is list

    if type(m0) is str:
        include.append(m0)
        m0 = get_mass(m0)

    if do_debug(debug):
        import time
        t0 = time.time()

    if type(exclude) is str:
        exclude = exclude.split(',')

    if fakefit:
        dofit = True

    if delta is None:
        delta0 = kargs.pop('delta0', .5)
        mask = (m >= m0 - delta0) * (m <= m0 + delta0)
        thresh = max(5, np.max(D[mask]) * kargs.get('min_perc', .01))
        li = np.argmax(mask) + np.argmax(D[mask] > thresh)
        ui = np.argmax(mask) + np.sum(mask) - np.argmax(D[mask][::-1] > thresh)
        upper_mass = m[ui] + .01
        lower_mass = m[li] - .01

        delta = (upper_mass - lower_mass) / 2.0 + .01
    else:
        lower_mass = m0 - delta
        upper_mass = m0 + delta
    if do_debug(debug):
        print("Lower mass: {:.2f}\nUpper mass: {:.2f}".format(lower_mass, upper_mass))
    mask = (m >= lower_mass) * (m <= upper_mass)

    negative = False
    if polarity in ['-', 'Negative', 'negative', 'neg', 'Neg', 'NEG']:
        negative = True
    if show_elts:
        if include_only is not None:
            if type(include_only) is str:
                E = include_only.split(",")
            else:
                E = include_only
        else:
            E = [x for NM in range(int(round(lower_mass)), int(round(upper_mass)) + 1) for x in
                 get_peaklist(NM, negative)]
            E = [x for x in E if x not in exclude] + include
        E = list(set([x + [['+', '-'][negative], '']['+' in x or '-' in x] for x in E]))
    else:
        E = []
    if negative:
        E = [x + ['-', '']['-' in x] for x in E]
    if do_debug(debug):
        print("Selected Elements: " + ", ".join(E))
    m0s = [get_mass(x) for x in E]
    E = [x for x, y in zip(E, m0s) if y >= lower_mass and y <= upper_mass]
    m0s = [get_mass(x) for x in E]
    if formula:
        E_labels = [formulafy(x) for x in E]
    else:
        E_labels = E
    if do_debug(debug):
        print("Displayed elements:", ", ".join(E))

    Dt = np.copy(D[mask])
    mt = m[mask]
    if do_debug(debug):
        print(" ; ".join(E))
    ax = kargs.pop('ax', plt.gca())
    if dofit or fakefit:
        mp = m[mask][np.argmax(D[mask])]  # mass of max peak
        dm = dm0
        if dm is None:
            dm = 0
        if len(E) > 0:
            i = np.argmin(abs(np.array([get_mass(x) for x in E if type(x) is str] + [x for x in E if type(
                x) is float]) - mp))  # which element is the closest to max_peak
            if dm0 is None:
                dm = mp - get_mass(E[i])
        p0 = [kargs.get('asym0', 1), dm] + [0, 0] * len(
            E)  # delta m is estimated from the deviation of the highest peak
        Et = copy.deepcopy(E)  # copy element list

        while len(Et) > 0:
            mp = mt[np.argmax(Dt)]  # mass of max peak
            ms = [get_mass(x) for x in Et]
            i = np.argmin(abs(ms - mp))
            idx = E.index(Et[i])
            j = np.argmin(abs(mt - ms[i] - dm))
            if do_debug(debug):
                print("Max element:", Et[i], mp, ms[i], Dt[j], idx)
            p0[2 + 2 * idx] = kargs.get('sig0', 0.002)
            p0[3 + 2 * idx] = Aredux * Dt[j]
            Dt -= LG(mt, ms[i], p0[2 + 2 * idx], amp=p0[3 + 2 * idx], asym=p0[0], lg=0)
            Dt[Dt < 0] = 0
            del Et[i]

        def fit(x, *p):
            y = x * 0
            for i in range((len(p) - 2) // 2):
                x0 = m0s[i] + p[1]
                y += LG(x, x0, p[2 + 2 * i], amp=p[3 + 2 * i], asym=p[0], lg=0)
            return y

        fit_type = None
        if dofit:
            try:
                assert not fakefit
                popt, pcov = curve_fit(fit, m[mask], D[mask], p0=p0,
                                       bounds=(
                                           [1 / kargs.get('asym_max', 10), -0.015] + [0, 0] * ((len(p0) - 1) // 2),
                                           [kargs.get('asym_max', 10), 0.015] + [kargs.get('sig_max', 0.01), np.inf] * (
                                                   (len(p0) - 1) // 2))
                                       )
                fit_type = 0
            except:
                p0[1] = 0
                try:
                    assert not fakefit
                    popt, pcov = curve_fit(fit, m[mask], D[mask], p0=p0,
                                           bounds=(
                                               [1 / kargs.get('asym_max', 4), -0.015] + [0, 0] * ((len(p0) - 1) // 2),
                                               [kargs.get('asym_max', 4), 0.015] + [kargs.get('sig_max', 0.01),
                                                                                    np.inf] * ((len(p0) - 1) // 2))
                                           )
                    fit_type = 1
                except Exception as e:
                    fit_type = 2
                    if do_debug(debug):
                        raise e
                    popt = p0
                    pcov = np.zeros((len(p0), len(p0)))
                    if ax is not None:
                        for x in ['right', 'left', 'top', 'bottom']:
                            ax.spines[x].set_color('red')
        else:
            popt = p0
            popt[1] = 0
            pcov = np.zeros((len(p0), len(p0)))
        if fakefit:
            popt[1] = dm0
        res = {}
        err = np.sqrt(np.diag(pcov))
        for i in range((len(popt) - 1) // 2):
            Y = LG(m[mask], m0s[i], popt[2 + 2 * i], popt[3 + 2 * i], lg=0, asym=popt[0])
            Area = popt[2 * i + 3] * popt[2 * i + 2] * np.sqrt(2 * np.pi) * (.5 + .5 * popt[0])
            Area_err = np.sqrt(
                (err[2 * i + 3] * popt[2 * i + 2] * np.sqrt(2 * np.pi) * (.5 + .5 * popt[0])) ** 2 +
                (err[2 * i + 2] * popt[2 * i + 3] * np.sqrt(2 * np.pi) * (.5 + .5 * popt[0])) ** 2 +
                (err[0] * popt[2 * i + 2] * popt[2 * i + 3] * np.sqrt(2 * np.pi) * .5) ** 2
            )
            if not errors:
                res[E_labels[i]] = {
                    'm0': m0s[i],
                    'mass': m0s[i] + popt[1],
                    'Area': Area,
                    'Amp': popt[2 * i + 3],
                    'sig': popt[2 * i + 2],
                    'assym': popt[0],
                    'dm': popt[1] * 1e6 / m0s[i],
                    'fit': fit_type
                }
            else:
                res[E_labels[i]] = {
                    'm0': m0s[i],
                    'mass': (m0s[i] + popt[1], err[1]),
                    'Area': (Area, Area_err),
                    'Amp': (popt[2 * i + 3], err[2 * i + 3]),
                    'sig': (popt[2 * i + 2], err[2 * i + 2]),
                    'assym': (popt[0], err[0]),
                    'dm': (popt[1] * 1e6 / m0s[i], err[1] * 1e6 / m0s[i]),
                    'fit': fit_type
                }
    else:
        popt = [0, 0]
        res = None
    if ax is not None:
        if label is None:
            p = ax.plot(m[mask] - popt[1], D[mask], color=kargs.get('curve_color', None))
        else:
            p = ax.plot(m[mask] - popt[1], D[mask], label=label, color=kargs.get('curve_color', None))

    if show_elts and ax is not None:
        from . import put_Xlabels
        if dofit and ax is not None:
            ax.plot(m[mask], D[mask], color=p[0].get_color(), alpha=.1)
            if pretty:
                put_Xlabels(ax, m0s, ["{0}: {res[Area]:.2f}".format(E, res=res[E]) for E in res], color=colors,
                            debug=dec_debug(debug), **kargs)
            resO = [(res[x]['m0'], res[x]) for x in res]
            resO.sort(key=lambda x: x[0])
            for i, (_, r) in enumerate(resO):
                col = colors[i % len(colors)]
                Y = LG(m[mask], r['m0'], r['sig'], r['Amp'], lg=0, asym=r['assym'])
                ax.plot(m[mask], Y, '--', color=col);
        elif pretty:
            put_Xlabels(ax, m0s, E_labels, color=colors, debug=dec_debug(debug), **kargs)

        if 'dsf' in kargs and 'dk0' in kargs and 'sf' in kargs and 'k0' in kargs:
            sf, dsf, k0, dk0 = [kargs[x] for x in 'sf,dsf,k0,dk0'.split(',')]
            P = list(zip(m0s, E))
            P.sort(key=lambda x: x[0])
            for i, (mi, Ei) in enumerate(P):
                col = colors[i % len(colors)]
                dmi = get_dm(mi, sf, k0, dsf, dk0)
                ax.axvline(mi - dmi, color=col, alpha=.5, linestyle=':')
                ax.axvline(mi + dmi, color=col, alpha=.5, linestyle=':')

        if not pretty:
            P = list(zip(m0s, E_labels))
            P.sort(key=lambda x: x[0])
            y = ax.get_ylim()[1]
            for i, (mi, Ei) in enumerate(P):
                col = colors[i % len(colors)]
                ax.axvline(mi, color=col, alpha=.5)
                ax.annotate(Ei, (mi, y), (5, 0), rotation=90, va='top', ha='left', textcoords='offset pixels',
                            color=col)
        if auto_scale:
            fig = plt.gcf()
            renderer = fig.canvas.get_renderer()
            ymax = ax.get_ylim()[1]
            ymin = ax.get_ylim()[0]
            for child in ax.get_children():
                bbox = child.get_window_extent(renderer)
                bbox_data = bbox.transformed(ax.transData.inverted())
                if ymax < bbox_data.ymax:
                    ymax = bbox_data.ymax
            ax.set_ylim((ymin, ymax))
            ax.autoscale_view()

            # Add the Intensity=0 line in case there is negative intensities (happens when plotting PCA components)
    if zero_axis:
        if type(zero_axis) is dict:
            ax.achline(0, **zero_axis)
        else:
            ax.axhline(0, color='k', alpha=.5, lw=.5)

    if dofit and ax is not None:
        ax.plot(m[mask] - popt[1], fit(m[mask], *popt), 'r:');
    return res


def plot_isotopes(elt, amp=None, main=None, ax=None, sig=None, asym=None, lg=0, limit=1, color='C1', show_elts=False,
                  debug=False, dm=0, **kargs):
    """
    plot the isotopes of a given element on a spectral profile plot
    """
    # Old parameter name compatibility
    if 'Amp' in kargs:
        from warnings import warn
        warn("Parameter Amp is deprecated. Please use amp in order to set the amplitude!")
        amp = kargs.pop("Amp")
    if 'showElts' in kargs:
        from warnings import warn
        warn("Parameter showElts is deprecated. Please use show_elts")
        show_elts = kargs.pop("showElts")
    from . import get_isotopes, LG, get_mass, get_abund, Molecule
    import matplotlib.pyplot as plt
    import numpy as np
    import re
    from scipy.optimize import curve_fit
    if type(elt) is Molecule:
        elt = str(elt)
    if ax is None:
        ax = plt.gca()
    if main is None:
        main = ax
    main_isotope = re.sub('^([0-9]+)', '', elt)
    main_iso = (main_isotope, get_mass(main_isotope), get_abund(main_isotope))
    L = main.lines[0]
    m, y = L.get_xdata(), L.get_ydata()
    m -= dm
    if hasattr(ax, 'log') and ax.log:
        y = 10 ** y
    mask = np.abs(m - main_iso[1]) < .1
    if amp is None:
        i = np.argmin(np.abs(m - main_iso[1]))
        amp = np.max(y[i - 10:i + 10]) / main_iso[2]
    if sig is None and asym is None:
        def fit(x, s, a):
            return LG(x, main_iso[1], s, amp=amp * main_iso[2], lg=lg, asym=a)

        (sig, asym), _ = curve_fit(fit, m[mask], y[mask], (0.005, 1))
    elif sig is None:
        def fit(x, s):
            return LG(x, main_iso[1], s, amp=amp * main_iso[2], lg=lg, asym=asym)

        s, _ = curve_fit(fit, m[mask], y[mask], (0.005))
        sig = s[0]
    elif asym is None:
        def fit(x, a):
            return LG(x, main_iso[1], sig, amp=amp * main_iso[2], lg=lg, asym=a)

        a, _ = curve_fit(fit, m[mask], y[mask], (1))
        asym = a[0]
    if debug:
        print(sig, asym)
    isos = get_isotopes(elt, min_abund=limit / amp)
    L = ax.lines[0]
    m, y = L.get_xdata(), L.get_ydata()
    if hasattr(ax, 'log') and ax.log:
        y = 10 ** y
    r = main.get_xlim()
    s = m * 0

    for iso in isos:
        s += LG(m, iso[1], sig, amp=amp * iso[2], lg=lg, asym=asym)
        if show_elts:
            ax.annotate(iso[0], (iso[1], amp * iso[2]), (0, 5), textcoords='offset pixels', ha='center')
    if hasattr(ax, 'log') and ax.log:
        s[s >= 1] = np.log10(s[s >= 1])
        s[s < 1] = 0
    ax.plot(m + dm, s, color, **kargs);
    return m, s



# -- coding: utf-8 --

# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

import base64
import os
import struct
import xml.etree.ElementTree as ET

import numpy as np

import pySPM.SPM
from pySPM.SPM import SPM_image, funit
from .utils.misc import aliased, alias, deprecated


@deprecated("getCurve")
def get_curve(filename, channel='Normal Deflection', backward=False):
    """
    function to retrieve data which are not in the form of images.
    This is typically used for 1D channel where the normal deflection is recorded while z is swept.
    """
    tree = ET.parse(filename)
    root = tree.getroot()
    namespace = {'spm': 'http://www.nanoscan.ch/SPM'}
    RAW = root.findall("spm:vector/spm:contents/spm:direction/spm:vector/"
                       "spm:contents/spm:name[spm:v='{direction}']/../spm:channel/spm:vector/"
                       "spm:contents/spm:name[spm:v='{channel}']/../spm:data/spm:v"
                       .format(direction=['forward', 'backward'][backward],
                               channel=channel), namespace)[0].text
    start = float(root.findall("spm:vector/spm:contents/spm:axis/spm:vector/"
                               "spm:contents/spm:start/spm:vector/spm:v", namespace)[0].text)
    stop = float(root.findall("spm:vector/spm:contents/spm:axis/spm:vector/"
                              "spm:contents/spm:stop/spm:vector/spm:v", namespace)[0].text)
    unit = root.findall("spm:vector/spm:contents/spm:axis/spm:vector/spm:contents"
                        "/spm:unit/spm:v", namespace)[0].text
    BIN = base64.b64decode(RAW)
    N = len(BIN)
    vals = np.array(struct.unpack("<" + str(N // 4) + "f", BIN))
    x = np.linspace(start, stop, len(vals))
    return x, vals


@aliased
class Nanoscan():
    def __init__(self, filename=None):
        if not os.path.exists(filename):
            raise IOError('File "{0}" Not Found'.format(filename))
        if filename[-4:] != '.xml':
            raise TypeError("Nanoscan files should be xml files")
        self.filename = filename
        tree = ET.parse(filename)
        self.root = tree.getroot()

        if self.root.tag == "{http://www.nanoscan.ch/SPM}scan":
            self.namespaces = {'spm': "http://www.nanoscan.ch/SPM"}
            self.type = "Nanoscan"
            self.fbPath = "spm:vector/spm:contents/spm:instrumental_parameters/spm:contents/spm:z_control/spm:contents"
            self.pixel_size = [int(z) for z in self.__grab(
                "spm:vector/spm:contents/spm:size/spm:contents//spm:v")]
            uval = float(self.__grab(
                ".//spm:area//spm:contents/spm:size/spm:contents/spm:fast_axis/spm:v"))
            udispu = self.__grab(
                ".//spm:area//spm:contents/spm:display_unit/spm:v")
            udisps = float(self.__grab(
                ".//spm:area/spm:contents/spm:display_scale/spm:v"))
            uname = self.__grab(".//spm:area/spm:contents/spm:unit/spm:v")
            x = funit(uval * udisps, udispu)
            uval = float(self.__grab(
                ".//spm:area//spm:contents/spm:size/spm:contents/spm:slow_axis/spm:v"))
            y = funit(uval * udisps, udispu)
            self.size = {
                'unit': x['unit'],
                'x': x['value'],
                'y': y['value']}
            try:
                self.feedback = {'channel': self.__grab(
                    '{0}/spm:z_feedback_channel/spm:v'.format(self.fbPath))}
                self.feedback['P'] = {
                    'value': float(self.__grab('{0}/spm:proportional_z_gain/spm:v'.format(self.fbPath))),
                    'unit': self.__grab('{0}/spm:proportional_z_gain_unit/spm:v'.format(self.fbPath))}
                self.feedback['I'] = {
                    'value': float(self.__grab('{0}/spm:integral_z_time/spm:v'.format(self.fbPath))),
                    'unit': self.__grab('{0}/spm:integral_z_time_unit/spm:v'.format(self.fbPath))}
                if self.feedback['channel'] == 'df':
                    self.feedback['channel'] = u'Δf'
            except:
                self.feedback = {}
            self.scan_speed = {
                z: {
                    'value': float(self.__grab(
                        "spm:vector//spm:direction/spm:vector/spm:contents/spm:name[spm:v='{dir}']/../spm:point_interval/spm:v".format(
                            dir=z))) * self.pixel_size[0],
                    'unit': self.__grab(
                        "spm:vector//spm:direction/spm:vector/spm:contents/spm:name[spm:v='{dir}']/../spm:point_interval_unit/spm:v".format(
                            dir=z))} for z in ['forward', 'backward']}
        else:
            raise TypeError(
                "Unknown or wrong data type. Expecting a valid Nanoscan xml")

    def list_channels(self):
        """
        Printout the list of stored channels
        """
        for d in ['forward', 'backward']:
            print(d)
            print("=" * len(d))
            for z in self.root.findall(
                    "spm:vector//spm:direction/spm:vector/spm:contents/spm:name[spm:v='{}']/../spm:channel//spm:contents/spm:name/spm:v".format(
                        d), self.namespaces):
                print("  - " + z.text)
            print()

    def get_channel(self, channel='Topography', backward=False, corr=None):
        try:
            RAW = self.__grab("spm:vector//spm:direction/spm:vector/spm:contents"
                              "/spm:name[spm:v='{direction}']/../spm:channel//spm:contents/spm:name[spm:v='{channel}']"
                              "/../spm:data/spm:v".format(direction=["forward", "backward"][backward], channel=channel))
        except:
            raise 'Channel {0} in {1} scan not found'.format(
                channel, "backward" if backward else "forward")
            return None

        BIN = base64.b64decode(RAW)
        recorded_length = len(BIN) / 4

        py = int(recorded_length / self.pixel_size[0])
        recorded_size = {
            'x': self.size['x'],
            'y': self.size['y'] * py / float(self.pixel_size[1]),
            'unit': self.size['unit']}

        image_array = np.array(struct.unpack("<%if" % (recorded_length), BIN)).reshape(
            (py, self.pixel_size[0]))
        return SPM_image(image_array, channel=channel, _type=self.type, real=recorded_size, corr=corr)

    def __grab(self, path):
        result = [z.text for z in self.root.findall(path, self.namespaces)]
        if len(result) == 1:
            result = result[0]
        return result

    @deprecated("arraySummary")
    def array_summary(self):
        res = [y.format(**self.__dict__) for y in
               ["{filename}", "{pixel_size[0]}×{pixel_size[1]}", "{size[x][value]}×{size[y][value]} {size[x][unit]}",
                "{scan_speed[forward][value]} {scan_speed[forward][unit]}",
                "{feedback[channel]}", "{P[value]:.2f} {P[unit]}", "{I[value]:.2f} {I[unit]}"]]

    @alias("getSummary")
    def get_summary(self):
        x = funit(self.size['x'], self.size['unit'])
        y = funit(self.size['y'], self.size['unit'])
        P = funit(self.feedback['P'])
        I = funit(self.feedback['I'])
        return u"""Feedback: {feedback[channel]} : P:{P[value]}{P[unit]} : I:{I[value]}{I[unit]}
Size: {pixel_size[0]}×{pixel_size[1]} pixels = {x[value]:.3} {x[unit]}×{y[value]:.3} {y[unit]}
Scan Speed: {scanSpeed[value]}{scanSpeed[unit]}/line""".format(
            x=x, y=y, P=P, I=I,
            feedback=self.feedback, pixel_size=self.pixel_size, size=self.size,
            scanSpeed=self.scan_speed['forward'])

    @staticmethod
    def show_dir_summary(path):
        from pySPM.utils import htmlTable
        res = [["Filename", "pixel size", "real size",
                "scan_speed", "feedback", "P", "I"]]
        for x in os.listdir(path):
            try:
                A = pySPM.Nanoscan(path + x)
                res.append([y.format(f=os.path.basename(A.filename), **A.__dict__) for y in
                            ["{f}", "{pixel_size[0]}×{pixel_size[1]}", "{size[x]}×{size[y]} {size[unit]}",
                             "{scan_speed[forward][value]} {scan_speed[forward][unit]}",
                             "{feedback[channel]}", "{feedback[P][value]:.2f} {feedback[P][unit]}",
                             "{feedback[I][value]:.2f} {feedback[I][unit]}"]])
            except:
                print("Cannot read image \"" + x + "\" skipping it")
        htmlTable(res, header=True)



from warnings import warn

import numpy as np


def dec_debug(debug):
    if debug > 0:
        return debug - 1
    if debug < 0:
        return debug + 1
    return False


def do_debug(debug):
    return debug == -1 or debug > 0


def adaptive(img):
    import skimage
    mi = np.min(img)
    img = np.asarray(256 ** 2 * (img - mi) / (np.max(img) - mi), dtype=np.uint16)
    img = skimage.exposure.equalize_adapthist(img, clip_limit=0.03)
    return img


def smiley(width, height=None, ratio=.9, eye=.15, thick=.1, eye_sep=.35, eye_height=.3, mouth_rad=.5, mouth_thick=.1):
    if height is None:
        height = width
    size = min(width, height)
    Y, X = np.mgrid[-height // 2:height // 2, -width // 2:width // 2]
    R = np.sqrt(X ** 2 + Y ** 2) / (size / 2)
    smiley = (R < ratio) * (R > ((ratio - thick)))
    smiley += np.sqrt(
        (X - eye_sep * ratio * size / 2) ** 2 + (Y + eye_height * ratio * size / 2) ** 2) < ratio * eye * size / 2
    smiley += np.sqrt(
        (X + eye_sep * ratio * size / 2) ** 2 + (Y + eye_height * ratio * size / 2) ** 2) < ratio * eye * size / 2
    smiley += (R < ratio * mouth_rad) * (R > (ratio * mouth_rad - ratio * mouth_thick)) * (Y > 0)
    return smiley * 1.0


import functools


def aliased(cls):
    original_methods = cls.__dict__.copy()
    for name in original_methods:
        method = original_methods[name]
        if hasattr(method, '_aliases'):
            for alias in method._aliases:
                setattr(cls, *alias)
    return cls


class alias(object):
    def __init__(self, *aliases):
        self.aliases = aliases

    def __call__(self, func):
        if not hasattr(func, '_aliases'):
            func._aliases = []
        func._aliases += [(x, func) for x in self.aliases]
        import sys
        module = sys.modules[func.__module__]
        for alias in self.aliases:
            setattr(module, alias, func)
        return func


class deprecated(object):
    def __init__(self, alias):
        self.alias = alias

    def __call__(self, func):
        import sys
        msg = "the function {} is DEPRECATED. Please use {} instead".format(self.alias, func.__name__)
        module = sys.modules[func.__module__]

        @functools.wraps(func)
        def wrapper(*args, **kargs):
            from warnings import warn
            warn(msg)
            return func(*args, **kargs)

        setattr(module, self.alias, wrapper)
        if not hasattr(func, '_aliases'):
            func._aliases = []
        func._aliases += [(self.alias, wrapper)]
        return func


@alias("get_bam", "get_BAM")
def getBAM(x, x0, N=10, least_one=False):
    """
    get the profile of the GaAlAs layer from a BAM-L200 sample.
    
    Parameters
    ----------
    x: numpy.ndarray
        x-axis in nm
    x0: float
        offset of the x-axis in nm (ie. where the first edge of the BAM is located)
    N: int
        The number of headings grating (seems to be 13 on the SEM image in the doc, but 10 in reality?)
    least_one: bool
        if True each GaAlAs strip are at least 1px wide.
    """
    P = x * 0
    l = 0

    # Here are the certified known distances
    d = dict(W1=691, W2=691, W3=293, W4=294, W5=19.5, W6=195, W7=195, W8=38, W9=3.6, W10=14.2, W11=3.5, W12=96, W13=5,
             W14=1, P1=587, P2=389, P3=273, P4=193, P5=136, P6=97, P7=67.5, P8=48.5, P9=76.5, P10=57, P11=42, P12=31,
             P13=23, P14=17.5, P15=13.3, P16=9.4, P17=6.9, P18=4.6, P19=3, P20=2)
    # The prelines
    lines = [(-147 * i, 80) for i in range(N, 0, -1)]

    # Estimated coordinates of the starting of lines
    pos = dict(W1=0, P1=d['W1'] + d['W2'], W5=2683, P9=8760, W8=7300, W9=7800, W10=8250, W11=9500, W12=9840, W14=8100)
    pos_rel = dict(W5=dict(P2=397, P3=1380, P4=2140, P5=2730, P6=3200, P7=3535, P8=3730),
                   P9=dict(P10=231, P11=402, P12=539, P13=642, P14=710),
                   W11=dict(P15=60, P16=104, P17=141, P18=162, P19=182, P20=191))
    for rel in pos_rel:
        for key in pos_rel[rel]:
            pos[key] = pos[rel] + pos_rel[rel][key]

    for key in sorted(pos):
        width = [d[key], d[key] // 2][key[0] == 'P']
        lines.append((pos[key], width))
        if key[0] == 'P': lines.append((pos[key] + d[key], width))

    for pos, width in lines:
        P[((x - x0) >= pos) * ((x - x0) <= pos + width)] = 1
        if least_one: P[np.argmin(abs(x - x0 - pos))] = 1  # At least 1 pixel set ?
    return P


def in_ipynb():
    try:
        # noinspection PyUnresolvedReferences
        cfg = get_ipython().config
        if 'IPKernelApp' in cfg:
            return True
        else:
            return False
    except NameError:
        return False


if in_ipynb():
    try:
        from tqdm.notebook import tqdm
    except:
        try:
            from tqdm import tqdm
        except:
            warn("the library tqdm cannot be found. All progressbar will be disabled.")
            tqdm = lambda x: x
    PB = tqdm
else:
    try:
        from tqdm import tqdm
    except:
        warn("the library tqdm cannot be found. All progressbar will be disabled.")
        tqdm = lambda x: x
    PB = tqdm



# -- coding: utf-8 --

# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

"""
This is an old module which is deprecated now. It is kept in the project as it can be used to read BIF6 and BIF3D images.
"""

import os
import re
import struct

import matplotlib.pyplot as plt
import numpy as np

import pySPM

Elements = {
    'H': {1: (1.00782503223, .999885), 2: (2.01410177812, .000115)},
    'Li': {6: (6.0151228874, .0759), 7: (7.0160034366, .9241)},
    'Be': {9: (9.012183065, 1)},
    'B': {10: (10.01293695, .199), 11: (11.00930536, .801)},
    'C': {12: (12, .9893), 13: (13.00335483507, .01107)},
    'N': {14: (14.00307400443, .99636), 15: (15.00010889888, .00364)},
    'O': {16: (15.99491461957, .99757), 17: (16.99913175650, .00038), 18: (17.99915961286, .00205)},
    'F': {19: (18.9984036273, 1)},
    'Ag': {107: (106.9050916, .51839), 109: (108.9047553, .48161)},
    'Au': {197: (196.96656879, 1)},
    'Si': {28: (27.97692653465, .92223), 29: (28.97649466490, .04685), 30: (29.973770136, 0.3092)},
    'Ti': {46: (45.95262772, .0825), 47: (46.95175879, 0.0744), 48: (47.94794198, .7372), 49: (48.94786568, .0541),
           50: (49.94478689, .0518)},
}


def getSpecElt(Elts):
    if len(Elts) == 1:
        return Elements[Elts[0]]
    tm = {}
    r1 = Elements[Elts[0]]
    r2 = getSpecElt(Elts[1:])
    for x in r1:
        for y in r2:
            if x + y not in tm:
                tm[x + y] = [r1[x][0] + r2[y][0], r1[x][1] * r2[y][1]]
            else:
                tm[x + y][1] += r1[x][1] * r2[y][1]
    return tm


def SplitElts(elt):
    elts = []
    for x, i in re.findall('([A-Z][a-z]?)([0-9]*)', elt):
        if i == '':
            i = 1
        else:
            i = int(i)
        elts += [x for k in range(i)]
    return elts


def showElts(Elts):
    r = getSpecElt(Elts)
    plt.bar(r.keys(), [z[1] for z in r.values()])
    plt.show()


class BIF6:
    def __init__(self, filename):
        self.f = open(filename, 'rb')
        self.header = struct.unpack('xx4s5H', self.f.read(16))
        self.size = (self.header[2], self.header[3])
        self.N = self.size[0] * self.size[1]
        self.cat = []
        for i in range(self.header[1]):
            self.f.seek(16 + i * (4 * self.N + 16))
            self.cat.append(struct.unpack('4f', self.f.read(16)))

    def getImgID(self, ID):
        assert ID >= 0 and ID <= self.header[1]
        self.f.seek(32 + ID * (4 * self.N + 16))
        return np.array(struct.unpack(str(self.N) + 'I', self.f.read(4 * self.N))).reshape(self.size)

    def getImgMass(self, masses, raw=False):
        if type(masses) is float or type(masses) is int:
            masses = [masses]
        SUM = None
        for i, x in enumerate(self.cat):
            for m in masses:
                if m >= x[0] - .5 and m <= x[1] + .5:
                    if SUM is None:
                        SUM = self.getImgID(i)
                    else:
                        SUM += self.getImgID(i)
        if raw:
            return SUM
        if SUM is None:
            return None
        return pySPM.SPM_image(np.flipud(SUM))

    def getImgElt(self, elt):
        r = {}
        A = getSpecElt(SplitElts(elt))
        for k in A:
            I = self.getImgMass(k)
            if I is not None:
                r[k] = {
                    'data': I,
                    'mass': A[k][0],
                    'abund': A[k][1]}
        return r

    def showImgElt(self, elt, size=10, abundCorr=False, tot=True):
        A = self.getImgElt(elt)
        ks = [z for z in A if A[z]['data'] != None]
        fig, ax = plt.subplots((len(ks) + 1) // 4 + 1, 4,
                               figsize=(10 * ((len(ks) + 1) // 4 + 1), 10))
        mi = np.min(A[ks[0]]['data'].pixels)
        ma = np.max(A[ks[0]]['data'].pixels)
        A[ks[0]]['CT'] = ma
        for i, k in enumerate(ks[1:]):
            M = np.min(A[k]['data'].pixels)
            mi = min(mi, M)
            A[k]['minCT'] = M
            M = np.max(A[k]['data'].pixels)
            ma = max(ma, M)
            A[k]['CT'] = M
        ks.sort()
        di = 0
        if tot:
            SUM = sum([A[k]['data'].pixels for k in ks])
            ax[0][0].imshow(SUM, vmin=0)
            ax[0][0].set_title("Total")
            di = 1
        for i, k in enumerate(ks):
            if abundCorr:
                ax[(i + di) // 4][(i + di) % 4].imshow(A[k]
                                                       ['data'].pixels / A[k]['abund'], vmin=0, vmax=ma)
            else:
                A[k]['data'].show(ax=ax[(i + di) // 4][(i + di) %
                                                       4], vmin=0, vmax=ma)
            ax[(i + di) // 4][(i + di) % 4].set_title(
                'mass: {mass:.3} - abundancy: {abund:.3f} - CT: {CT}'.format(**A[k]))

    def __exit__(self, exc_type, exc_value, traceback):
        self.f.close()


class BIF3D:

    def __init__(self, path):
        self.Peaks = {}
        self.RPeaks = {}
        self.Path = os.path.dirname(path)
        self.Basename = os.path.basename(path)
        self.size = None
        for x in os.listdir(self.Path):
            if x[:len(self.Basename)] == self.Basename:
                if x[-6:] == ".BIF3D":
                    s = self.Basename + r' \(([0-9]+)\)(?: - (.*?))?\.BIF3D'
                    r = re.search(s, x)
                    ID = int(r.group(1))
                    self.Peaks[ID] = r.group(2)
                    if r.group(2) != None:
                        self.RPeaks[r.group(2)] = ID
        self.data = {}

    def getIDs(self, channels):
        assert type(channels) in [str, int, list, tuple]
        if type(channels) == int:
            return channels
        if type(channels) == str:
            assert channels in self.RPeaks.keys()
            return self.RPeaks[channels]
        if type(channels) == list or type(channels) == tuple:
            ID = []
            for x in channels:
                ID.append(self.getIDs(x))
            return ID

    def getChannel(self, channel):
        k = self.getIDs(channel)
        if k in self.data:
            return self.data[k]
        v = self.Peaks[k]
        if v != None:
            path = '{path}{sep}{basename} ({ID}) - {Peak}.BIF3D'
        else:
            path = '{path}{sep}{basename} ({ID}).BIF3D'
        f = open(path.format(path=self.Path, sep=os.sep,
                             basename=self.Basename, ID=k, Peak=v), 'rb')
        A = f.read()
        f.close()
        size = struct.unpack("II", A[32:40])
        if self.size == None:
            self.size = size
        else:
            assert self.size == size
        return np.array(struct.unpack("{0}d".format(self.size[0] * self.size[1]), A[640:])).reshape(self.size)

    def loadChannel(self, channel):
        k = self.getIDs(channel)
        if not k in self.data:
            self.data[k] = self.getChannel(channel)

    def loadChannels(self, channels):
        if not type(channels) in [list, tuple]:
            self.loadChannel(channels)
        else:
            for x in channels:
                self.loadChannel(x)

    def getChannels(self, *channels):
        k = self.getIDs(channels)
        self.loadChannels(k)
        if type(k) == int:
            return self.data[k]
        D = np.zeros(self.size)
        for x in k:
            D += self.data[x]
        return D

    def listChannels(self):
        return self.Peaks

    def show(self, ax, *channels):
        D = self.getChannels(*channels)
        ax.imshow(D)
        if len(channels) > 1:
            ax.set_title(','.join([str(z) for z in channels]))
        else:
            ax.set_title(str(channels[0]))



# -- coding: utf-8 --

# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

from __future__ import absolute_import

from . import align, utils
from .Bruker import Bruker
from .ITA import ITA, ITA_collection
from .ITAX import ITAX
from .ITM import ITM
from .ITS import ITS
from .SPM import *
from .SXM import SXM
from ._version import __version__
from .collection import Collection
from .nanoscan import Nanoscan
from .utils import constants as const

__all__ = ["ITA", "ITAX", "ITS", "ITM", "PCA", "Block", "SPM", "Bruker", "nanoscan", "utils", "SXM"]



# -- coding: utf-8 --

# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

"""
This is a standalone script which allows the user to perfome cross-section on ToF-SIMS images on different channels
"""

import os
import sys

import numpy as np
from PyQt5.QtCore import QCoreApplication
from PyQt5.QtWidgets import QApplication, QTableWidgetItem, QFileDialog, QWidget, QAction
from scipy import optimize as opt

import pySPM
from pySPM.utils import CDF
from slicer import Ui_slicer


class SlicerApp(QWidget):
    def __init__(self, filename=None):
        super(QWidget, self).__init__()
        self.ui = Ui_slicer()
        self.ui.setupUi(self)

        self.canvas = self.ui.mpl.canvas
        self.fig = self.ui.mpl.canvas.fig
        self.initPlotLayout()
        self.level = None
        if filename is None:
            if len(sys.argv) < 2:
                self.path, _ = QFileDialog.getOpenFileName(self, "ITA Image", "", "(*.ITA)")
            else:
                # If an argument is sent to the script, the first argument will be used as a Path. Very usefull for debugging the script without having to selectr the folder each time with window dialog
                self.path = sys.argv[1]
        else:
            self.path = filename
        if not os.path.exists(self.path):
            raise Exception("File \"{}\" is not found".format(self.path))
        if os.path.exists(self.path + ".level.npy"):
            self.level = np.load(self.path + ".level.npy")
        self.curs = [0, 0, 0]
        self.volume = None
        self.ITA = pySPM.ITA(self.path)
        for i, x in enumerate(self.ITA.get_masses()):
            self.ui.peakList.setRowCount(i + 1)
            self.ui.peakList.setItem(i, 0, QTableWidgetItem(x['assign']))
            self.ui.peakList.setItem(i, 1, QTableWidgetItem("{:.2f}u".format((x['cmass']))))
            self.ui.peakList.setItem(i, 2, QTableWidgetItem("{:.2f}u".format(x['umass'] - x['lmass'])))
        self.ui.peakList.show()
        self.ui.cmap.currentIndexChanged.connect(self.plot)
        self.ui.prof1daxis.currentIndexChanged.connect(self.plot)
        self.ui.peakList.cellClicked.connect(self.load_channel)
        self.canvas.mpl_connect('button_press_event', self.on_pick)
        self.flatAction = QAction("Flatten substrate from this channel")
        self.flatAction.triggered.connect(self.flatten)
        self.ui.correction.stateChanged.connect(self.plot)
        self.ui.peakList.addAction(self.flatAction)
        self.ui.status.setText("IDLE")

    def flatline(self, y):
        for x in range(self.volume.shape[1]):
            popt, pcov = opt.curve_fit(CDF, np.arange(self.volume.shape[2]), self.volume[y, x, :],
                                       (10, self.volume.shape[2] / 2, 1))
            self.level[y, x] = popt[1]

    def flatten(self):
        self.ui.status.setText("Start the flattening...")
        self.level = np.zeros(self.volume.shape[:2])
        self.ui.pb.setMaximum(self.volume.shape[0])
        for y in range(self.volume.shape[0]):
            self.ui.pb.setValue(y)
            self.flatline(y)
            QCoreApplication.processEvents()
            self.ax.clear()
            self.ax.imshow(self.level)
            self.canvas.draw()
        self.ui.pb.setValue(0)
        self.ui.status.setText("Flattening finished")
        np.save(self.path + ".level", self.level)

    def load_channel(self, row, col):
        self.ui.status.setText("Loading channel...")
        id = row
        vol = []
        for i in range(self.ITA.Nscan):
            x = self.ITA.getImage(id, i)
            vol.append(x)
        self.volume = np.stack([x for x in vol], axis=2)
        if not self.level is None:
            self.corrected = np.zeros(self.volume.shape)
            z = np.arange(self.ITA.Nscan)
            self.ui.pb.setMaximum(self.level.shape[0])
            for y in np.arange(self.level.shape[0]):
                self.ui.pb.setValue(y)
                for x in np.arange(self.level.shape[1]):
                    dz = int(-self.level[y, x] + np.max(self.level))
                    self.corrected[y, x, dz:] = self.volume[y, x, :self.volume.shape[2] - dz]
            self.ui.pb.setValue(0)
        self.plot()
        self.ui.status.setText("IDLE")

    def plot(self):
        if self.ui.correction.isChecked():
            A = self.corrected
        else:
            A = self.volume
        cmap = self.ui.cmap.currentText()
        self.axXY.clear()
        self.axXZ.clear()
        self.axYZ.clear()
        self.ax.clear()
        if self.volume is None:
            return
        self.axXY.imshow(A[:, :, self.curs[2]], cmap=cmap)
        self.axXZ.imshow(A[self.curs[1], :, :].T, cmap=cmap)
        self.axYZ.imshow(A[:, self.curs[0], :].T, cmap=cmap)
        self.axXY.axhline(self.curs[1])
        self.axXY.axvline(self.curs[0])
        self.axXZ.axhline(self.curs[2])
        self.axXZ.axvline(self.curs[0])
        self.axYZ.axhline(self.curs[2])
        self.axYZ.axvline(self.curs[1])
        self.axXY.set_title("XY")
        self.axXZ.set_title("XZ")
        self.axYZ.set_title("YZ")
        if self.ui.prof1daxis.currentText() in 'XYZ':
            i = 'XYZ'.index(self.ui.prof1daxis.currentText())
            self.ax.set_xlabel("XYZ"[i])
            if i == 0:
                self.ax.plot(A[:, self.curs[0], self.curs[2]])
            elif i == 1:
                self.ax.plot(A[self.curs[1], :, self.curs[2]])
            elif i == 2:
                self.ax.plot(A[self.curs[1], self.curs[0], :])
        self.canvas.draw()

    def on_pick(self, event):
        if not event.inaxes in [self.axYZ, self.axXZ, self.axXY]:
            return
        x = event.xdata
        y = event.ydata
        axis = [self.axYZ, self.axXZ, self.axXY].index(event.inaxes)
        xdata = int(x)
        ydata = int(y)

        if event.inaxes == self.axXY:
            self.curs[0] = xdata
            self.curs[1] = ydata
        elif event.inaxes == self.axYZ:
            self.curs[1] = xdata
            self.curs[2] = ydata
        elif event.inaxes == self.axXZ:
            self.curs[0] = xdata
            self.curs[2] = ydata
        else:
            print("Click event not handled")
        self.curs = [np.clip(0, self.curs[i], self.volume.shape[i] - 1) for i in range(3)]
        self.plot()

    def initPlotLayout(self):
        """
        Setup the plotting layout.
        """
        self.axXY = self.fig.add_subplot(2, 2, 1)
        self.axYZ = self.fig.add_subplot(2, 2, 2)
        self.axXZ = self.fig.add_subplot(2, 2, 3)
        self.ax = self.fig.add_subplot(2, 2, 4)
        self.fig.tight_layout()


app = QApplication(sys.argv)
window = SlicerApp()
window.show()
sys.exit(app.exec_())



# -- coding: utf-8 --

# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

import os.path

from pySPM import Block


class InvalidRAWdataformat(Exception):
    def __init__(self, block, msg):
        self.block = block
        self.msg = msg

    def __str__(self):
        return "Invalid RAW dataformat seen in block " + self.block.parent + '/' + self.block.name + ' : ' + self.msg


class ITS:
    def __init__(self, filename, debug=False):
        """
        ITS
        """
        self.filename = filename
        assert os.path.exists(filename)
        self.f = open(self.filename, 'rb')
        self.Type = self.f.read(8)
        assert self.Type == b'ITStrF01'
        self.root = Block.Block(self.f)



# -- coding: utf-8 --

# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

"""
Module to handle SPM images recorded by a Bruker AFM
"""

import struct
import re
import struct

import numpy as np

import pySPM


class Bruker:
    """
    Class to handle SPM images recorded by a Bruker AFM
    """

    def __init__(self, path):
        self.path = path
        self.file = open(self.path, 'rb')
        self.layers = []
        self.scanners = []
        mode = ''
        while True:
            line = self.file.readline().rstrip().replace(b'\\', b"")
            if line == b'*Ciao image list':
                self.layers.append({})
                mode = 'Image'
            elif line == b'*Scanner list':
                self.scanners.append({})
                mode = 'Scanner'
            else:
                args = line.split(b": ")
                if len(args) > 1:
                    if mode == 'Image':
                        self.layers[-1][args[0]] = args[1:]
                    elif mode == 'Scanner':
                        self.scanners[-1][args[0]] = args[1:]
                if line == b"*File list end":
                    break

    def _get_bpp(self, i):
        off = int(self.layers[i][b'Data offset'][0])
        cols = int(self.layers[i][b'Number of lines'][0])
        rows = int(self.layers[i][b'Samps/line'][0])
        byte_length = int(self.layers[i][b'Data length'][0])
        length = rows * cols
        bpp = byte_length // length
        return bpp

    def _get_raw_layer(self, i, debug=False):
        """
        Internal function to retrieve raw data of a layer
        """
        off = int(self.layers[i][b'Data offset'][0])
        if debug:
            print("RAW offset: ", off)
        cols = int(self.layers[i][b'Number of lines'][0])
        rows = int(self.layers[i][b'Samps/line'][0])
        byte_length = int(self.layers[i][b'Data length'][0])
        length = rows * cols
        bpp = byte_length // length
        byte_length = length * bpp

        self.file.seek(off)
        return np.array(
            struct.unpack("<" + str(length) + {2: 'h', 4: 'i', 8: 'q'}[bpp], self.file.read(byte_length)),
            dtype='float64').reshape((cols, rows))

    def list_channels(self, encoding='latin1'):
        print("Channels")
        print("========")

        for layer in self.layers:
            try:
                print(layer[b'@2:Image Data'][0].decode(encoding))
            except KeyError:
                pass
        for layer in self.layers:
            try:
                print(layer[b'@3:Image Data'][0].decode(encoding) + " (MFM)")
            except KeyError:
                pass

    def get_channel(self, channel="Height Sensor", backward=False, corr=None, debug=False, encoding='latin1', lazy=True,
                    mfm=False):
        """
        Load the SPM image contained in a channel
        """
        for i in range(len(self.layers)):
            if mfm:
                # Handle case of MFM, where image data is stored at layer 3 on the backward channel
                backward = True
                _type = "Bruker MFM"
                try:
                    layer_name = self.layers[i][b'@3:Image Data'][0].decode(encoding)
                except KeyError:
                    continue
            else:
                _type = "Bruker AFM"
                try:
                    layer_name = self.layers[i][b'@2:Image Data'][0].decode(encoding)
                except KeyError:
                    continue
            result = re.match(
                r'([^ ]+) \[([^\]]*)\] "([^"]*)"', layer_name).groups()
            if result[2] == channel:
                if debug:
                    print("channel " + channel + " Found!")
                bck = False
                if self.layers[i][b'Line Direction'][0] == b'Retrace':
                    bck = True
                if bck == backward:
                    if debug:
                        print("Direction found")
                    var = self.layers[i][b'@2:Z scale'][0].decode(encoding)
                    if debug:
                        print("@2:Z scale", var)
                    if '[' in var:
                        result = re.match(r'[A-Z]+\s+\[([^\]]+)\]\s+\(-?[0-9\.]+ .*?\)\s+(-?[0-9\.]+)\s+(.*?)$',
                                          var).groups()
                        if debug:
                            print(result)
                        bpp = int(self.layers[i][b'Bytes/pixel'][0])
                        if debug:
                            print("BPP", bpp)
                        # scale = float(result[1])
                        scale = float(result[1]) / 256 ** bpp

                        result2 = self.scanners[0][b'@' + result[0].encode(encoding)][0].split()
                        if debug:
                            print("result2", result2)
                        scale2 = float(result2[1])
                        if len(result2) > 2:
                            zscale = result2[2]
                        else:
                            zscale = result2[0]
                        if b'/V' in zscale:
                            zscale = zscale.replace(b'/V', b'')
                        if debug:
                            print("scale: {:.3e}".format(scale))
                            print("scale2: {:.3e}".format(scale2))
                            print("zscale: " + str(zscale))
                        var = self.layers[i][b'@2:Z offset'][0].decode(encoding)
                        result = re.match(r'[A-Z]+\s+\[[^\]]+\]\s+\(-?[0-9\.]+ .*?\)\s+(-?[0-9\.]+)\s+.*?$',
                                          var).groups()
                        offset = float(result[0])
                    else:
                        if debug:
                            print("mode 2")
                        result = re.match(r'[A-Z]+ \(-?[0-9\.]+ [^\)]+\)\s+(-?[0-9\.]+) [\w]+', var).groups()
                        scale = float(result[0]) / 65536.0
                        scale2 = 1
                        zscale = b'V'
                        result = re.match(r'[A-Z]+ \(-?[0-9\.]+ .*?\)\s+(-?[0-9\.]+) .*?',
                                          self.layers[i][b'@2:Z offset'][0].decode(encoding)).groups()
                        offset = float(result[0])
                    if debug:
                        print("Offset:", offset)
                    data = self._get_raw_layer(i, debug=debug) * scale * scale2
                    xres = int(self.layers[i][b"Samps/line"][0])
                    yres = int(self.layers[i][b"Number of lines"][0])
                    if debug:
                        print("xres/yres", xres, yres)
                    scan_size = self.layers[i][b'Scan Size'][0].split()
                    aspect_ratio = [int(x) for x in self.layers[i][b'Aspect Ratio'][0].split(b":")]
                    if debug:
                        print("aspect ratio", aspect_ratio)
                    if scan_size[2][0] == 126:
                        scan_size[2] = b'u' + scan_size[2][1:]
                    size = {
                        'x': float(scan_size[0]) / aspect_ratio[1],
                        'y': float(scan_size[1]) / aspect_ratio[0],
                        'unit': scan_size[2].decode(encoding)}
                    image = pySPM.SPM_image(
                        channel=[channel, 'Topography'][channel == 'Height Sensor'],
                        BIN=data,
                        real=size,
                        _type=_type,
                        zscale=zscale.decode(encoding),
                        corr=corr)
                    return image
        if lazy:
            return self.get_channel(channel=channel, backward=not backward, corr=corr, debug=debug, encoding=encoding,
                                    lazy=False)
        raise Exception("Channel {} not found".format(channel))



# -- coding: utf-8 --

# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

"""
This module handle sxm file format used by Nanonis instruments
"""

import os
import struct

import numpy as np

from .SPM import SPM_image


# -- coding: utf-8 --
# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>


class SXM:
    def __init__(self, filename):
        self.filename = filename
        assert os.path.exists(self.filename)
        self.f = open(self.filename, 'rb')
        l = ''
        key = ''
        self.header = {}
        while l != b':SCANIT_END:':
            l = self.f.readline().rstrip()
            if l[:1] == b':':
                key = l.split(b':')[1].decode('ascii')
                self.header[key] = []
            else:
                if l:  # remove empty lines
                    try:
                        self.header[key].append(l.decode('ascii').split())
                    except UnicodeDecodeError:
                        self.header[key].append(l.decode('unicode_escape').split())

        while self.f.read(1) != b'\x1a':
            pass
        assert self.f.read(1) == b'\x04'
        assert self.header['SCANIT_TYPE'][0][0] in ['FLOAT', 'INT', 'UINT', 'DOUBLE']
        self.data_offset = self.f.tell()
        self.f.close()
        self.size = dict(pixels={
            'x': int(self.header['SCAN_PIXELS'][0][0]),
            'y': int(self.header['SCAN_PIXELS'][0][1])
        }, real={
            'x': float(self.header['SCAN_RANGE'][0][0]),
            'y': float(self.header['SCAN_RANGE'][0][1]),
            'unit': 'm'
        })

    def list_channels(self):
        print("Channels")
        print("========")
        h = self.header['DATA_INFO'][0]
        i = h.index('Name')
        for z in self.header['DATA_INFO'][1:]:
            print("  - " + z[i])

    def get_channel(self, name, direction='forward', corr=None):
        chID = 0
        zscale = ''
        for x in self.header['DATA_INFO'][1:]:
            if x[1] == name:
                if x[3] == 'both' and direction == 'backward':
                    chID += 1
                if x[3] == 'both' or direction == x[3]:
                    break
                return None
                zscale = x[2]
            elif x[3] == 'both':
                chID += 2
            else:
                chID += 1

        size = self.size['pixels']['x'] * self.size['pixels']['y']
        self.f = open(self.filename, 'rb')
        self.f.seek(self.data_offset + chID * size * 4)
        data = np.array(struct.unpack('<>'['MSBFIRST' == self.header['SCANIT_TYPE'][0][1]] + str(size) +
                                      {'FLOAT': 'f', 'INT': 'i', 'UINT': 'I', 'DOUBLE': 'd'}[
                                          self.header['SCANIT_TYPE'][0][0]], self.f.read(4 * size))).reshape(
            (self.size['pixels']['y'], self.size['pixels']['x']))
        self.f.close()
        return SPM_image(
            channel=name,
            BIN=data,
            real=self.size['real'],
            _type='Nanonis SXM',
            zscale=zscale,
            corr=corr)



import numpy as np

from . import constants as const

_SI_units = ['kg', 'm', 's', 'A', 'K', 'cd', 'mol']

_units = {
    'V': {'kg': 1, 'm': 2, 's': -3, 'A': -1, 'K': 0, 'cd': 0, 'mol': 0},
    'C': {'kg': 0, 'm': 0, 's': 1, 'A': 1, 'K': 0, 'cd': 0, 'mol': 0},
    'N': {'kg': 1, 'm': 1, 's': -2, 'A': 0, 'K': 0, 'cd': 0, 'mol': 0},
    'J': {'kg': 1, 'm': 2, 's': -2, 'A': 0, 'K': 0, 'cd': 0, 'mol': 0},
    'W': {'kg': 1, 'm': 2, 's': -3, 'A': 0, 'K': 0, 'cd': 0, 'mol': 0},
    'Pa': {'kg': 1, 'm': -1, 's': -2, 'A': 0, 'K': 0, 'cd': 0, 'mol': 0},
    'Ω': {'kg': 1, 'm': 2, 's': -3, 'A': -2, 'K': 0, 'cd': 0, 'mol': 0}
}

_units_scale_conversion = {
    'eV': (const.qe, 'V'),
    'bar': (1e5, 'Pa'),
    'atm': (101325, 'Pa')
}

_unit_shift_conversion = {
    '°C': (273.15, 'K'),
}

_unit_scale = {
    'y': -24, 'z': -21, 'a': -18, 'f': -15, 'p': -12, 'n': -9, 'u': -6, 'm': -3, 'c': -2, 'd': -1,
    'h': 2, 'k': 3, 'M': 6, 'G': 9, 'T': 12, 'P': 15, 'E': 18, 'Z': 21, 'Y': 24, '': 0}

_unit_revert_scale = {}
import re

_SI_units = ['kg', 'm', 's', 'A', 'K', 'cd', 'mol']

_units = {
    'V': {'kg': 1, 'm': 2, 's': -3, 'A': -1},
    'C': {'s': 1, 'A': 1},
    'N': {'kg': 1, 'm': 1, 's': -2},
    'J': {'kg': 1, 'm': 2, 's': -2},
    'W': {'kg': 1, 'm': 2, 's': -3},
    'Pa': {'kg': 1, 'm': -1, 's': -2},
}

_unit_scale = {
    'y': -24, 'z': -21, 'a': -18, 'f': -15, 'p': -12, 'n': -9, 'u': -6, 'm': -3, 'c': -2, 'd': -1,
    'h': 2, 'k': 3, 'M': 6, 'G': 9, 'T': 12, 'P': 15, 'E': 18, 'Z': 21, 'Y': 24, '': 0}


def par_parse(s):
    parsed = []
    count = 0
    opening = None
    closing = 0
    for i, x in enumerate(s):
        if x is '(':
            if opening is None:
                opening = i
            count += 1
        elif x is ')':
            count -= 1
            if count == 0 and opening is not None:
                parsed += [s[closing:opening], par_parse(s[opening + 1:i])]
                closing = i + 1
                opening = None
    if closing < len(s):
        parsed.append(s[closing:])
    return parsed


def op_parse(s):
    r = []
    for x in s:
        if type(x) is list:
            r.append(op_parse(x))
        else:
            r += [x for x in re.split(r'(\*|/|\^)', x) if len(x) > 0]
    return r


def parse(unit):
    sp = par_parse(unit)
    sp = op_parse(sp)
    sp = u_parse(sp)
    sp = op_exec(sp)
    return sp


def num_parse(s):
    pass


def u_parse(s):
    if type(s) is list:
        sub = [u_parse(y) for y in s]
        return sub
    for x in '*/^':
        if x in s:
            return s
    result = None
    if re.match(r'\-?[0-9]+(\.[0-9]+)?', s):
        result = unit({}, float(s))
    elif s in _SI_units:
        result = unit(s)
    elif s in _units:
        result = unit(_units[s])
    elif s[0] in _unit_scale:
        if s[1:] in _SI_units:
            result = unit(s[1:], 10 ** (_unit_scale[s[0]]))
        elif s[1:] in _units:
            result = unit(_units[s[1:]], 10 ** (_unit_scale[s[0]]))
    elif len(s) == 2 and s[1] == 'g' and x[0] in _unit_scale:
        result = unit('kg', 10 ** (_unit_scale[s[0]] - 3))
    elif s == 'g':
        result = unit('kg', 1e-3)
    elif s in _units_scale_conversion:
        u = _units_scale_conversion[s]
        result = unit(u[1], u[0])
    return result


def op_exec(s):
    s = [op_exec(x) if type(x) is list else x for x in s]
    while '^' in s:
        i = s.index('^')
        a = s[i - 1]
        b = s[i + 1]
        s = s[:i - 1] + [a ** b] + s[i + 2:]
    while '/' in s:
        i = s.index('/')
        s = s[:i - 1] + [s[i - 1] / s[i + 1]] + s[i + 2:]
    while '*' in s:
        i = s.index('*')
        s = s[:i - 1] + [s[i - 1] * s[i + 1]] + s[i + 2:]
    return s[0]


class unit(object):
    def __init__(self, u, value=1):
        self.value = 1
        if type(u) is str:
            if u in _SI_units:
                self.units = {u: 1}
            else:
                p = parse(u)
                self.units = p.units
                self.value = p.value
        elif type(u) is dict:
            self.units = {x: u[x] for x in u}
        else:
            raise TypeError(type(u))
        self.value *= value

    def __mul__(self, b):
        value = self.value
        units = self.units
        if isinstance(b, unit):
            for x in b.units:
                units[x] = units.get(x, 0) + b.units[x]
            value *= b.value
            return unit(units, value)
        return unit(self.units, self.value * b)

    __rmul__ = __mul__

    def __div__(self, b):
        value = self.value
        units = self.units
        if isinstance(b, unit):
            for x in b.units:
                units[x] = units.get(x, 0) - b.units[x]
            value /= b.value
            return unit(units, value)
        return unit(self.units, self.value / b)

    def __rdiv__(self, b):
        value = 1 / self.value
        units = {x: -self.units[x] for x in self.units}
        if isinstance(b, unit):
            for x in b.units:
                units[x] = units.get(x, 0) + b.units[x]
            value *= b.value
            return unit(units, value)
        return unit(units, b / self.value)

    def __pow__(self, n):
        if isinstance(n, unit):
            assert n.units == {}
            return unit({x: n.value * self.units[x] for x in self.units}, self.value ** n.value)
        return unit({x: n * x for x in self.units}, self.value ** n)

    __truediv__ = __div__
    __rtruediv__ = __rdiv__

    def __repr__(self):
        if self.units == {}:
            return str(self.value)
        u = '*'.join(['{}^{}'.format(x, self.units[x]) if self.units[x] is not 1 else x for x in self.units if
                      self.units[x] != 0])
        if self.value == 1:
            return u
        return '{:.3e}*'.format(self.value) + u

    __str__ = __repr__


class SIunit(np.ndarray):
    def __new__(cls, input_array, u={}):
        obj = np.asarray(input_array).view(cls)
        obj.unit = unit(u)
        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # see InfoArray.__array_finalize__ for comments
        if obj is None: return
        self.unit = getattr(obj, 'unit', {})

    def __repr__(self):
        return np.ndarray.__repr__(self)[:-1] + ', unit=' + str(self.unit) + ')'

    def __mul__(self, b):
        r = np.ndarray.__mul__(self, b)
        if isinstance(b, SIunit):
            r.unit = r.unit * b.unit
        else:
            r.unit = self.unit
        return r

    def __add__(self, b):
        r = np.ndarray.__add__(self, b)
        if isinstance(b, SIunit):
            for x in self.units:
                assert r.units[x] == b.units[x]
            for x in b.units:
                assert r.units[x] == b.units[x]
            return r
